import numpy as np
import pandas as pd
from operator import itemgetter
import math
import warnings
warnings.filterwarnings('ignore')

import os
os.environ["CUDA_VISIBLE_DEVICES"]="0"

#Dataset
file_path_full_training_set = '../Data/KDDTrain+.txt'
file_path_test = '../Data/KDDTest+.txt'
# file_path_test = '../Data/KDDTest-21.txt'

columns = (['duration','protocol_type','service','flag','src_bytes','dst_bytes','land','wrong_fragment','urgent','hot','num_failed_logins','logged_in'
            ,'num_compromised','root_shell','su_attempted','num_root','num_file_creations','num_shells','num_access_files','num_outbound_cmds','is_host_login'
            ,'is_guest_login','count','srv_count','serror_rate','srv_serror_rate','rerror_rate','srv_rerror_rate','same_srv_rate','diff_srv_rate'
            ,'srv_diff_host_rate','dst_host_count','dst_host_srv_count','dst_host_same_srv_rate','dst_host_diff_srv_rate','dst_host_same_src_port_rate'
            ,'dst_host_srv_diff_host_rate','dst_host_serror_rate','dst_host_srv_serror_rate','dst_host_rerror_rate','dst_host_srv_rerror_rate','attack_class','level'])

#Load data
df = pd.read_csv(file_path_full_training_set, header=None)
test_df = pd.read_csv(file_path_test, header=None)

df.columns = columns
test_df.columns = columns

df=df.drop('level',axis=1)
test_df=test_df.drop('level',axis=1)

#Data process
#Label Encoder
#Dos:1 Probe:2 U2R:3 R2L:4
Dos_type = ['back','land','neptune','pod','smurf','teardrop','mailbomb','processtable','udpstorm','apache2','worm']
Probe_type = ['ipsweep','nmap','portsweep','satan','mscan','saint']
U2R_type = ['buffer_overflow','loadmodule','perl','rootkit','xterm','ps','sqlattack','httptunnel']
R2L_type = ['ftp_write','guess_passwd','imap','multihop','phf','warezmaster','warezclient','spy','sendmail','xlock','snmpguess','named','xsnoop','snmpgetattack']

attack_labels = ['Normal','DoS','Probe','U2R','R2L']

def class_attack(attack_class):
    if attack_class in Dos_type:
        attack_type = 1
    elif attack_class in Probe_type:
        attack_type = 2
    elif attack_class in U2R_type:
        attack_type = 3
    elif attack_class in R2L_type:
        attack_type = 4
    else:
        attack_type = 0
    return attack_type

attack_class = df.attack_class.apply(class_attack)
df['attack_class'] = attack_class

test_attack_class = test_df.attack_class.apply(class_attack)
test_df['attack_class'] = test_attack_class

# print(df['attack_class'].value_counts())
# print(test_df['attack_class'].value_counts())

#Convert non numerical features to numerical features
def map_encoding(df,cols):
    new_df = df.copy()
    col_dicts = {}
    for each_col in cols:
        data = new_df.drop_duplicates(subset=[each_col],keep='first',inplace=False)
        col_value = data[each_col].values
        tmp={}
        count = 1
        for k in col_value:
            tmp[k] = count
            count +=1
        col_dicts[each_col]=tmp
        new_df[each_col] = new_df[each_col].map(col_dicts[each_col])
    return new_df

#Feature scaling,Min-Max scaler
def normalize(df, cols):
    """
    @param df pandas DataFrame
    @param cols a list of columns to encode
    @return a DataFrame with normalized specified features
    """
    result = df.copy() # do not touch the original df
    for feature_name in cols:
        max_value = df[feature_name].max()
        min_value = df[feature_name].min()
        if max_value > min_value:
            result[feature_name] = (df[feature_name] - min_value) / (max_value - min_value)
    return result

#Calculate the information gain of discrete features
def IG_dis(df):
    list1 = ['protocol_type', 'service', 'flag']
    info = []
    for k in range(len(list1)):
        pro = df[list1[k]].unique()
        count1 = list(0 for i in range(len(pro)))
        count2 = list(0 for i in range(len(pro)))
        # df_list = list(df)
        for i in range(len(df)):
            for j in range(len(pro)):
                if (df.loc[i, list1[k]] == pro[j]):
                    if (df.loc[i, 'attack_class'] == 0):
                        count1[j] = count1[j] + 1
                    else:
                        count2[j] = count2[j] + 1

        # Record the conditional information entropy of label category 0 and label category 1
        ent = list(0.0 for i in range(len(pro)))
        for i in range(len(ent)):
            x = count1[i] * 1.0 / (count1[i] + count2[i])
            if x != 0.0:
                ent[i] = x * math.log(x)
            if x != 1.0:
                ent[i] -= (1 - x) * math.log(1 - x)
        sub = 0.0
        for i in range(len(ent)):
            sub += (count1[i] + count2[i]) * 1.0 / len(df) * ent[i]

        h = -(sum(count1) / len(df)) * math.log(sum(count1) / len(df))
        h += -(sum(count2) / len(df)) * math.log(sum(count2) / len(df))
        info_enhance = h - sub

        info.append(info_enhance)

    return info

#Calculate the information gain of continuous features
class IG_con():
    def __init__(self, X, y):
        X = np.array(X)
        n_feature = X.shape[1]
        n_y = len(y)

        orig_H = 0
        for i in set(y):
            orig_H += -(y.count(i) / n_y) * math.log(y.count(i) / n_y)#Calculate the original information entropy

        condi_H_list = []
        for i in range(n_feature):
            feature = X[:, i]
            sourted_feature = sorted(feature)
            threshold = [(sourted_feature[inde - 1] + sourted_feature[inde]) / 2 for inde in range(len(feature)) if
                         inde != 0]

            thre_set = set(threshold)
            if float(max(feature)) in thre_set:
                thre_set.remove(float(max(feature)))
            if min(feature) in thre_set:
                thre_set.remove(min(feature))
            pre_H = 0
            for thre in thre_set:
                lower = [y[s] for s in range(len(feature)) if feature[s] < thre]
                highter = [y[s] for s in range(len(feature)) if feature[s] > thre]
                H_l = 0
                for l in set(lower):
                    H_l += -(lower.count(l) / len(lower)) * math.log(lower.count(l) / len(lower))
                    # print(H_l)
                H_h = 0
                for h in set(highter):
                    H_h += -(highter.count(h) / len(highter)) * math.log(highter.count(h) / len(highter))
                    # print(H_h)
                temp_condi_H = len(lower) / n_y * H_l + len(highter) / n_y * H_h
                condi_H = orig_H - temp_condi_H#information gain
                # print(condi_H)
                pre_H = max(pre_H, condi_H)#the maximum information gain
                # print(pre_H)
            condi_H_list.append(pre_H)

        self.IG = condi_H_list

    def getIG(self):
        return self.IG

#Calculate the overall information gain
def info_gain(new_data):
    new_train = new_data.loc[:125972]
    # the information gain of discrete features
    dis_data = new_train
    temp1 = IG_dis(dis_data)
    print(temp1)
    # the information gain of continuous features
    con_data = new_train
    drop_cols = con_data.loc[:, [i for i in list(con_data.columns) if i in ['protocol_type', 'service', 'flag']]]
    con_data.drop(drop_cols, axis=1, inplace=True)
    con_data_X = con_data.drop('attack_class', axis=1)
    con_data_y = con_data['attack_class']
    con_data_y = con_data_y.replace([1], [1])
    con_data_y = con_data_y.replace([2], [1])
    con_data_y = con_data_y.replace([3], [1])
    con_data_y = con_data_y.replace([4], [1])
    con_data_y = list(con_data_y)
    temp2 = IG_con(con_data_X, con_data_y).getIG()
    print(temp2)
    temp = temp1 + temp2
    colus1=['protocol_type', 'service', 'flag']
    colus2=list(con_data_X.columns)
    colus=colus1+colus2
    # sorted_index = sorted(col_index, key=lambda x: IG_data[col_index.index(x)], reverse=True)
    temp, colus = (list(x) for x in zip(*sorted(zip(temp, colus),key=itemgetter(0),reverse=True)))

    #Save calculated information gain
    f = open("../Restructure_data/feature_info.txt", "w")
    for line in temp:
        f.write(str(line) + '\n')
    f.close()

    #Save sorted feature sequences
    f = open("../Restructure_data/feature_sort.txt", "w")
    for line in colus:
        f.write(str(line) + '\n')
    f.close()

    return colus

#Data Augmentation
from imblearn.over_sampling import ADASYN
def data_enhance(df):
    Normal = df[df.attack_class == 0]
    DDoS = df[df.attack_class == 1]
    PProbe = df[df.attack_class == 2]
    UU2R = df[df.attack_class == 3]
    RR2L = df[df.attack_class == 4]

    frames = [Normal, DDoS, PProbe, UU2R, RR2L]
    total_data = pd.concat(frames)

    oversample = ADASYN(sampling_strategy='auto',n_neighbors=2)
    data = total_data.copy()
    trainX = data.drop('attack_class', axis=1)
    trainY = data['attack_class']
    trainX, trainY = oversample.fit_resample(trainX, trainY)
    train_data = pd.concat([trainX, trainY], axis=1)
    print(train_data.shape)

    return train_data

new_data=pd.concat([df,test_df],axis=0,ignore_index=True)
new_data_X=new_data.drop('attack_class',axis=1)
new_data_y=new_data['attack_class']
# print(new_data.shape)

#Convert non numerical features to numerical features
cols1 = ['protocol_type', 'service', 'flag']
new_data=map_encoding(new_data,cols1)

columns1=columns[:-2]
new_data=normalize(new_data,columns1)

#Divide the training and testing sets
new_df=new_data.loc[:125972]
new_test_df=new_data.loc[125973:]

#Save preprocessed training and testing sets
# new_df.to_csv('../Restructure_data/train_data+.csv',index=False)
# new_test_df.to_csv('../Restructure_data/test_data+.csv',index=False)

total_data=pd.concat([new_df,new_test_df],axis=0)
total_y=total_data['attack_class']

#Data augmentation
enhance_data=data_enhance(new_df)
print(enhance_data.shape)
#Save the data with sample augmentation
# enhance_data.to_csv('../Restructure_data/train_data+_ADASYN.csv',index=False)

#Calculate information gain
feature_sort=info_gain(new_data)
sort_data=pd.DataFrame()
for i in range(len(feature_sort)):
    sort_data=sort_data.append(total_data[feature_sort[i]])

#sort feature
sort_data=pd.DataFrame(sort_data.values.T,columns = sort_data.index,index =sort_data.columns)
sort_data=pd.concat([sort_data,total_y],axis=1)

#Sorted training and testing sets
new_dff=sort_data.loc[:125972]
new_test_dff=sort_data.loc[125973:]
new_dff=pd.DataFrame(new_dff)
new_test_dff=pd.DataFrame(new_test_dff)
print(new_dff.shape)
print(new_test_dff.shape)

#Dataset with feature sort
new_dff.to_csv('../Restructure_data/train_data+_sort.csv',index=False)
new_test_dff.to_csv('../Restructure_data/test_data+_sort.csv',index=False)


